/*
 * Video implementation for Linux framebuffer devices
 *
 * This source code does not limit on 80 column screens, based on the assumption that
 * whoever wish to edit or see this code, use bigger framebuffer screens theirselves too.
 *
 */

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <signal.h>

#include <linux/vt.h>
#include <linux/fb.h>

#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>

#include "video.h"
#include "cons.h"

/* Permanent field - required for correctly removing the timered items */
static int Field[24][40];
static int Color[24][40];

/* Currently visible field - required for optimized output */
static int VField[24][40];
static int VColor[24][40];

/* Requested visible field - readable by other modules too */
int QField[24][40];
int QColor[24][40];

/* The current tty handle */
static int tty;

/* Framebuffer filehandle */
static int fb_fh;

#define XRES 640
#define YRES 480

static struct fb_var_screeninfo OldScr;

static int Active;
static void SetSwitcher(void);
static void Switcher(int signal)
{
	if(signal==SIGUSR1)
	{
		ioctl(tty, VT_RELDISP, 1);
		Active=0;
	}
	if(signal==SIGUSR2)
	{
		ioctl(tty, VT_RELDISP, VT_ACKACQ);
		Active=1;
		raise(SIGWINCH);
	}
	
	SetSwitcher();
}
static void SetSwitcher(void)
{
	signal(SIGUSR1, Switcher);
	signal(SIGUSR2, Switcher);
}


static int ch2char(int c)
{
	static char Mrks[16] = "#\"-'-`-^.|,<.>v+";
	                     /* 0 123456789ABCDEF          */
	                     /* &1=up, &2=lt, &4=rt, &8=dn */
		
	if(!iscorner(c))
	{
		/* We don't print confusing chars. */
		if(c < ' ')c = ' ';
		return c;
	}
	
	return Mrks[uncorner(c)];
}

void ClrScr(void)
{
	int x, y;
	
	for(y=0; y<24; y++)
		for(x=0; x<40; x++)
		{
			Field[y][x] = ' ';
			Color[y][x] = 7;
			
			QField[y][x] = ' ';
			QColor[y][x] = 7;
		}
}

void RefreshScreen(int x)
{
	int y;
	
	GetWinSize();
	
	for(y=0; y<24; y++)
		for(x=0; x<40; x++)
		{
			VField[y][x] = ' ';
			VColor[y][x] = -1;
		}
}

static int ReadMode(FILE *fp, char *line, struct fb_var_screeninfo *fb_var)
{
	int geometry = 0;
	int timings  = 0;
	
    /* fill in new values */
    fb_var->sync  = 0;
    fb_var->vmode = 0;
    
    while(fgets(line,79,fp) && !strstr(line,"endmode"))
    {
    	char value[16];
    	
        if(sscanf(line," geometry %d %d %d %d %d",
                  &fb_var->xres,
                  &fb_var->yres,
                  &fb_var->xres_virtual,
                  &fb_var->yres_virtual,
                  &fb_var->bits_per_pixel) == 5)
			geometry = 1;
        	
        if(sscanf(line," timings %ld %ld %ld %ld %ld %ld %ld",
                  &fb_var->pixclock,
                  &fb_var->left_margin,  &fb_var->right_margin,
                  &fb_var->upper_margin, &fb_var->lower_margin,
                  &fb_var->hsync_len,    &fb_var->vsync_len) == 7)
			timings = 1;
            
        if(sscanf(line, " hsync %15s",value)  ==1 && !strcasecmp(value,"high"))fb_var->sync |= FB_SYNC_HOR_HIGH_ACT;
        if(sscanf(line, " vsync %15s",value)  ==1 && !strcasecmp(value,"high"))fb_var->sync |= FB_SYNC_VERT_HIGH_ACT;
        if(sscanf(line, " csync %15s",value)  ==1 && !strcasecmp(value,"high"))fb_var->sync |= FB_SYNC_COMP_HIGH_ACT;
        if(sscanf(line, " extsync %15s",value)==1 && !strcasecmp(value,"true"))fb_var->sync |= FB_SYNC_EXT;
        if(sscanf(line, " laced %15s",value)  ==1 && !strcasecmp(value,"true"))fb_var->vmode |= FB_VMODE_INTERLACED;
        if(sscanf(line, " double %15s",value) ==1 && !strcasecmp(value,"true"))fb_var->vmode |= FB_VMODE_DOUBLE;
    }
    
    return !geometry || !timings;	/* Failure flag */
}

void InitVideo(void)
{
	struct vt_mode vt_mode;
	char line[80];
	char best[32];
	FILE *fp;
	long by,bx;
	
	InitCons();
	
    if(isatty(0))
    	tty = 0;
    else if((tty = open("/dev/tty",O_RDWR)) < 0)
    {
    	perror("/dev/tty");
		DoneCons();
    	exit(errno);
    }
   	
	if(ioctl(tty,VT_GETMODE, &vt_mode) < 0)
	{	
		perror("VT_GETMODE");
		DoneCons();
		exit(errno);
	}
    
    vt_mode.mode   = VT_PROCESS;
    vt_mode.waitv  = 0;
    vt_mode.relsig = SIGUSR1;
    vt_mode.acqsig = SIGUSR2;

	if(ioctl(tty,VT_SETMODE, &vt_mode) < 0)
	{
		perror("VT_SETMODE");
		DoneCons();
		exit(errno);
	}
	if((fb_fh=open("/dev/fb0", O_WRONLY)) < 0)
	{
		perror("/dev/fb0");
		DoneCons();
		exit(errno);
	}
	if(ioctl(fb_fh, FBIOGET_VSCREENINFO, &OldScr) < 0)
	{
		perror("FBIOGET_VSCREENINFO");
		DoneCons();
		exit(errno);
	}
	
    if(!(fp = fopen("/etc/fb.modes","r")))
    {
		perror("/etc/fb.modes");
		DoneCons();
		exit(errno);
	}    	
	
	bx=1000000;
	by=1000000;
	
    while(fgets(line,79,fp))
	{
        char label[32];
        if(sscanf(line, "mode \"%31[^\"]\"", label) == 1)
        {
			struct fb_var_screeninfo fb_var = OldScr;
			int xd, yd;
			
            if(ReadMode(fp, line, &fb_var))continue;
            
            if(fb_var.yres_virtual < YRES
            || fb_var.xres_virtual != XRES
            || fb_var.bits_per_pixel != 8)continue;
            
            xd = fb_var.xres-XRES;
            yd = fb_var.yres-YRES;
            
            if(xd*xd + yd*yd < bx+by)
            {
            	xd=bx*bx;
            	yd=by*by;
            	strcpy(best, label);
    }	}	}
    
    rewind(fp);
    while(fgets(line,79,fp))
	{
        char label[32];
        if(sscanf(line, "mode \"%31[^\"]\"", label) == 1
        && !strcmp(label, best))
        {
			struct fb_var_screeninfo fb_var = OldScr;
			
            if(ReadMode(fp, line, &fb_var))continue;
            
            bx = fb_var.xres;
            by = fb_var.yres;
            
            /* set */
            fb_var.xoffset = fb_var.yoffset = 0;
            if(ioctl(fb_fh, FBIOPUT_VSCREENINFO, &fb_var) < 0)
                perror("FBIOPUT_VSCREENINFO");
    }	}
    
	fprintf(stderr, "8-bit mode %dx%d, \"%s\" set\n", (int)bx, (int)by, best);

	ClrScr();
	
	#ifdef SIGWINCH
	signal(SIGWINCH, RefreshScreen);
	#endif
	
	SetSwitcher();
		
	RefreshScreen(0);
}

/* transp: 0=not transparent - 1=transparent - 2=over nonpermanent also
 *   time: 0=permanent       - other=disappears after this count of ticks
 *     ch: ascii character or corner combination
 *  color: combination of 1=blue, 2=green, 4=red, 8=intense
 */
void SetChar(int x, int y, int ch, int color, int transp, int time)
{
	/* Check against bounds */
	if(x<0 || x>39 || y<0 || y>23)return;
	
	ch=ch;
	color=color;
	transp=transp;
	time=time;
}

static void FlushScreen(void)
{
}
static void Evaluate(void)
{
}

void UpdateVideo(int numticks)
{
	while(numticks)
	{
		FlushScreen();
	
		/* After the screen has been draw,
		 * prepare already for the next frame..
		 */
		Evaluate();
		 
		numticks--;
	}
}

void DoneVideo(void)
{
	DoneCons();
    if(ioctl(fb_fh, FBIOPUT_VSCREENINFO, &OldScr) < 0)
		perror("FBIOPUT_VSCREENINFO");
}
